from django.contrib.auth import get_user_model
from rest_framework.serializers import CharField, DateTimeField

from console_base.api.serializers import (
    LCModelSerializer,
    LCHyperlinkedIdentityField,
    CanonicalIDField,
)

from encipher.models import EncryptionKey
from encipher.utils import generate_kid

BASE_READ_ONLY_FIELDS = (
    'pk',
    'url',
    'created',
    'modified',
)

User = get_user_model()


# ---------------------------------------------------------------------------
class EncryptionKeySerializer(LCModelSerializer):

    url = LCHyperlinkedIdentityField(view_name='api-encryptionkey-detail')
    user = CanonicalIDField(queryset=User.objects.all())
    kid = CharField(allow_null=True, allow_blank=True)
    revocation_date = DateTimeField(allow_null=True)

    class Meta:
        model = EncryptionKey
        fields = (
            'pk',
            'url',
            'created',
            'modified',
            'cid',
            'is_active',
            'user',
            'name',
            'key',
            'kid',
            'status',
            'last_used_on',
            'revocation_date',
        )
        read_only_fields = BASE_READ_ONLY_FIELDS

    def validate(self, attrs):
        """
        Ensure the KID value is present so the UniqueConstraint
        validation on the `kid` field succeeds.
        """
        if not attrs.get('kid'):
            attrs['kid'] = generate_kid(attrs['key'])
        return attrs


# ---------------------------------------------------------------------------
class EncryptionKeySerializerLite(LCModelSerializer):
    """
    Serializer with the minimal fields required for creation of Encryption Keys.
    Used for initial registering of new Encryption Keys from devices.
    """

    url = LCHyperlinkedIdentityField(view_name='api-encryptionkey-detail')
    user = CanonicalIDField(queryset=User.objects.all(), required=False, allow_null=True)

    class Meta:
        model = EncryptionKey
        fields = (
            'pk',
            'url',
            'created',
            'modified',
            'cid',
            'name',
            'key',
            'user',
        )
        read_only_fields = BASE_READ_ONLY_FIELDS


# ---------------------------------------------------------------------------
class EncryptionKeyDataTables(LCModelSerializer):

    class Meta:
        model = EncryptionKey
        fields = (
            'pk',
            'url',
            'created',
            'modified',
            'cid',
            'is_active',
            'user',
            'name',
            'kid',
            'status',
            'last_used_on',
            'revocation_date',
        )
        read_only_fields = BASE_READ_ONLY_FIELDS

__all__ = (
    'EncryptionKeySerializer',
    'EncryptionKeyDataTables',
    'EncryptionKeySerializerLite',
)
