from datetime import datetime
from lcconfig import ConsoleSettingsConfig
from lclazy import LazyLoader
import logging
from pathlib import Path
from typing import Optional, NamedTuple, TYPE_CHECKING
from uuid import UUID

from django.conf import settings
from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from rest_framework.exceptions import AuthenticationFailed

if TYPE_CHECKING:
    from lcrequests.typehints import PasetoFooter
    from .models import EncryptionKey
    import pylogcabin
    from uuid import UUID
else:
    pylogcabin = LazyLoader('pylogcabin', globals(), 'pylogcabin')

from .key_loaders import (
    get_key_password,
    KeyDetails,
    load_pinned_verification_key as base_load_pinned_verification_key,
    load_local_umbrella_api_key_from_file as base_load_local_umbrella_api_key_from_file,
)
from .utils import generate_kid

logger = logging.getLogger(__name__)
User = get_user_model()

ENCRYPTION_SECTION = 'Encryption'
LOCAL_UMBRELLA_KEY_NAME = 'local_umbrella_key_cid'


class VerificationKeyData(NamedTuple):
    """
    Canonical ID and Key of Private Key to sign Paseto Tokens.
    """
    pk: 'UUID'
    key: 'UUID'
    last_used_on: datetime


def register_key_on_aspen(public_key: 'EncryptionKey', api=None) -> bool:
    """
    Helper method to create the EncryptionKey on the Aspen CA server.
    """
    aspen_api = api or pylogcabin.AspenAPI(signed_token_class=load_aspen_token_class())

    if confirm_key_on_aspen(public_key, api):
        return True

    aspen_api.load('encipher', 'encryptionkey')  # noqa

    data = {
        'cid': str(public_key.cid),
        'user': str(public_key.user.cid),
        'key': public_key.key,
        'name': public_key.name,
        'revocation_date': None,
        'kid': generate_kid(public_key.key),
    }
    try:
        resp = aspen_api.post(post_data=data)
        return resp.ok
    except pylogcabin.exceptions.ConsoleAPIClientError as e:
        logger.error(f'Failed to create Console API key for {public_key}. Error: {e.response.text}')
    except Exception as e:
        logger.exception(e)
    return False


def confirm_key_on_aspen(public_key: 'EncryptionKey', api=None) -> bool:
    """
    Helper method to confirm that the EncryptionKey exists on the Aspen CA server.
    """
    aspen_api = api or pylogcabin.AspenAPI(signed_token_class=load_aspen_token_class())
    aspen_api.load('aspentls', 'verificationkey')  # noqa

    try:
        return aspen_api.head(cid=public_key.cid).ok
    except pylogcabin.exceptions.ConsoleAPIAuthenticationError:
        msg = 'Invalid credentials; unable to retrieve Verification Key from Aspen.'
        logger.error(msg)
        raise AuthenticationFailed(_(msg))
    except pylogcabin.exceptions.LCHTTPError as e:
        logger.info('An error occurred while retrieving credentials: %s', e)
        raise AuthenticationFailed(_('An error occurred while retrieving credentials'))


def retrieve_key_from_aspen(footer: 'PasetoFooter', api=None) -> VerificationKeyData:
    """
    The Aspen CA Server is the central repository for all Verification Keys.
    When a token references a kid or cid that isn't in our database, then
    attempt to retrieve it from Aspen to perform the validation.
    """
    aspen_api = api or pylogcabin.AspenAPI(signed_token_class=load_aspen_token_class())
    aspen_api.load('aspentls', 'verificationkey')  # noqa

    query_params = {'cid': footer['cid']}

    try:
        results = aspen_api.list(params=query_params).get('results', [])
        if len(results) > 1:
            AuthenticationFailed(_('Multiple verification keys returned'))
        return persist_aspen_key(results[0])

    except pylogcabin.exceptions.ConsoleAPIAuthenticationError:
        msg = 'Invalid credentials; unable to retrieve Verification Key from Aspen.'
        logger.error(msg)
        raise AuthenticationFailed(_(msg))
    except pylogcabin.exceptions.LCHTTPError as e:
        logger.info('An error occurred while retrieving credentials: %s', e)
        raise AuthenticationFailed(_('An error occurred while retrieving credentials'))


def persist_aspen_key(key_data: dict) -> VerificationKeyData:
    """
    Save key data from Aspen as new Encryption Key record.

    {
        "pk": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
        "url": "string",
        "created": "2023-12-18T21:27:00.623Z",
        "modified": "2023-12-18T21:27:00.623Z",
        "cid": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
        "is_active": true,
        "user": "3fa85f64-5717-4562-b3fc-2c963f66afa6",
        "name": "string",
        "key": "string",
        "kid": "string",
        "status": "active",
        "last_used_on": "2023-12-18T21:27:00.623Z",
        "revocation_date": "2023-12-18T21:27:00.623Z"
    }
    """
    from .api.serializers import EncryptionKeySerializer

    key = EncryptionKeySerializer(data=key_data)

    if not key.is_valid(raise_exception=True):
        logger.info(f'Unable to verify key data: {key_data}')
        raise AuthenticationFailed(_('Unable to verify key data from Aspen'))

    key = key.save()

    return VerificationKeyData(
        pk=key.pk,
        key=key.key,
        last_used_on=key.last_used_on,
    )


def load_pinned_verification_key() -> KeyDetails:
    """
    Wrapper for base_load_pinned_verification_key, which should be usable
    without initializing Django.

    Portals have "pinned" key baked into the installation configuration file,
    used for registration purposes.

    Since it's included in the installation config, it cannot be secured via
    PRIVATE_KEY_PASSWORD_PREFIX machinery

    This key should _not_ be used after registration has been completed.
    """
    try:
        return base_load_pinned_verification_key(
            settings.PINNED_VERIFICATION_KEY_CID,
            settings.PINNED_VERIFICATION_KEY_USER.cid,
            f'{settings.CONF_DIR}/{settings.PINNED_VERIFICATION_KEY_FILE}',
        )
    except ValueError:
        raise AuthenticationFailed(_('Unable to load pinned verification key')) from None


def load_local_umbrella_api_key_from_file() -> KeyDetails:
    """
    Wrapper for base_load_pinned_verification_key, which should be usable
    without initializing Django.

    The local Encryption Key should be loadable from a file,
    so that scripts can use it without initializing the whole
    Portal to make a database call.
    """
    try:
        return base_load_local_umbrella_api_key_from_file(
            settings.UMBRELLA_API_USER_CID,
            f'{settings.CONF_DIR}/{settings.PINNED_VERIFICATION_KEY_FILE}'
        )
    except ValueError:
        raise AuthenticationFailed(_('Local verification key does not exist')) from None


def save_local_umbrella_api_key_to_file(key: 'EncryptionKey') -> None:
    """
    Save local Encryption Key to a file for by scripts without
    initializing the whole Portal to query the database.
    """
    cfg = ConsoleSettingsConfig()
    cfg.save_section(ENCRYPTION_SECTION, data={LOCAL_UMBRELLA_KEY_NAME: str(key.cid)})
    password = get_key_password(key.cid, settings.PRIVATE_KEY_PASSWORD_PREFIX, ssh=True)

    with open(f'{settings.CONF_DIR}/{settings.LOCAL_VERIFICATION_KEY_FILE}', 'wb') as kf:
        kf.write(key.get_private_key().private_bytes(password))


def local_umbrella_key_registered_with_aspen() -> bool:
    """
    Local Umbrella API Key must be saved to disk and
    registered with Aspen before signing API tokens.

    The file should only exist on disk after it's been
    registered with Aspen.
    """
    local_key_file = Path(f'{settings.CONF_DIR}/{settings.LOCAL_VERIFICATION_KEY_FILE}')
    return local_key_file.exists() and local_key_file.stat().st_size > 0


def load_local_umbrella_api_key() -> Optional[KeyDetails]:
    """
    After initial registration has been performed,
    the local Encryption Key should be used for
    all Umbrella API communications.
    """
    if (
        not (user_cid := settings.UMBRELLA_API_USER_CID)
        or not local_umbrella_key_registered_with_aspen()
    ):
        return

    from .models import EncryptionKey

    key = EncryptionKey.objects.filter(
        name=settings.UMBRELLA_API_KEY_NAME,
        user__cid=user_cid,
    ).select_related('user').first()

    return KeyDetails(key.cid, key.user.cid, key.get_private_key()) if key else None


def load_aspen_token_class():
    """
    Generate a token for Aspen API communications,
    preferring to sign it with the Local Encryption Key.
    """
    if not (key_data := load_local_umbrella_api_key()):
        key_data = load_pinned_verification_key()

    return pylogcabin.AspenAPI.TOKEN_CLASS(
        private_key=key_data.key,
        user_cid=key_data.user_cid,
        verification_cid=key_data.cid,
    )
