"""
Provides RestFramework auth backend for signed tokens.
"""
try:
    import pybase64 as _base64
except (ModuleNotFoundError, ImportError):
    import base64 as _base64
import binascii
import logging
from lclazy import LazyLoader
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    import pyseto
else:
    pyseto = LazyLoader('pyseto', globals(), 'pyseto')

from rest_framework import exceptions
from rest_framework.authentication import get_authorization_header, BaseAuthentication
from django.conf import settings
from django.contrib.auth import get_user_model
from django.utils.translation import gettext_lazy as _
from rest_framework.request import Request

from encipher import get_setting
from encipher.tokens import AccessToken, UnverifiedToken, UnverifiedRegistrationToken
from encipher.typehints import VerifiedAuth
from encipher.utils import lookup_user_record

logger = logging.getLogger(__name__)
User = get_user_model()


# ---------------------------------------------------------------------------
class PasetoTokenMixin:
    """
    Paseto token authentication mixin for Django Rest Framework.
    """

    header_method_key = ''

    def authenticate(self, request: Request):
        # Ensure this auth header was meant for us (it has the Paseto Token auth method).
        auth = get_authorization_header(request).split()

        method = self.authenticate_header(request)
        if not auth or auth[0] != method.encode():
            return None

        try:

            header_token = auth[1].decode()
            return self.authenticate_credentials(header_token, request)

        except IndexError:
            error_msg = _('Invalid signed token header. No credentials provided.')
        except UnicodeError:
            error_msg = _('Invalid signed token header. Token string contains invalid characters.')

        raise exceptions.AuthenticationFailed(error_msg)

    def lookup_user_record(self, token: AccessToken, request: Request) -> VerifiedAuth:
        """
        Lookup user record from token claims
        """
        user = lookup_user_record(token)
        if not user:
            raise exceptions.AuthenticationFailed(_('User not found for provided token.'))
        if not user.is_active:
            raise exceptions.AuthenticationFailed(_('User inactive or deleted.'))

        return VerifiedAuth(user, token)

    def authenticate_header(self, request: Request) -> str:
        if not self.header_method_key:
            return ''
        return get_setting(self.header_method_key)

    def authenticate_credentials(self, token: str, request):
        raise NotImplementedError('Override on subclasses')


# ---------------------------------------------------------------------------
class PasetoAuthentication(PasetoTokenMixin, BaseAuthentication):
    """
    Clients should authenticate by passing the token key in the "Authorization"
    HTTP header, prepended with the string "Paseto ".  For example:

        Authorization: Paseto 401f7ac837da42b97f613d789819ff93537bee6a
    """
    header_method_key = 'AUTH_METHOD'

    def authenticate_credentials(self, token: str, request: Request) -> VerifiedAuth:
        """
        Authenticate user by comparing signed token with any
        EncryptionKey records linked to the specified user.
        """
        try:

            access_token = UnverifiedToken(token).verify()
            access_token.verify()
            return self.lookup_user_record(access_token, request)

        except (ValueError, pyseto.PysetoError) as e:
            error = str(e).lower()
            msg = 'Invalid token error: %s'
            if 'token expired' in error or 'invalid exp' in error or 'invalid nbf' in error:
                msg = f'{msg} - check time on both client and server'
            logger.error(msg, e)
            raise exceptions.AuthenticationFailed(_(msg) % e) from None


# ---------------------------------------------------------------------------
class PasetoAutoAuthentication(PasetoTokenMixin, BaseAuthentication):
    """
    This is a DANGEROUS operation, and should only be done in specific
    situations, such as on pre-defined acceptable end-points, and when
    the user has no other EncryptionKey records defined.

    Validates Paseto tokens for Django Rest Framework, to be used
    in registration situations, such as the creation of the FIRST
    EncryptionKey for a user.

    The public key value must be included in the POST body or GET query param
    with the `verify_public_key` field.
    (The public key value  must be base64 encoded when provided via the GET query param.)

        Authorization: Auto_Paseto 401f7ac837da42b97f613d789819ff93537bee6a
    """
    header_method_key = 'AUTO_AUTH_METHOD'

    def authenticate_credentials(self, token: str, request: Request) -> VerifiedAuth:
        """
        Get verification public key from `request.data`, POST body or GET query params,
        in that order. GET param must be base64-encoded.
        """
        # Allow HEAD method so requests can query whether the key exists or not.
        if request.method not in ('POST', 'PUT', 'PATCH', 'HEAD'):
            raise exceptions.AuthenticationFailed('Invalid registration method')

        if (
                not request.resolver_match
                or request.resolver_match.url_name
                not in settings.AUTO_REGISTER_ENCRYPTION_KEY_URL_NAMES
        ):
            raise exceptions.AuthenticationFailed(_('Invalid registration URL'))

        try:

            key_pem = self.load_public_key(request)
            access_token = UnverifiedRegistrationToken(token, pem=key_pem).verify()
            access_token.verify()
            return self.lookup_user_record(access_token, request)

        except (ValueError, pyseto.PysetoError) as e:
            logger.error('Invalid token error: %s', e)
            raise exceptions.AuthenticationFailed(_('Invalid token error: %s') % e) from None

    def lookup_user_record(self, token: AccessToken, request: Request) -> VerifiedAuth:
        """
        Prevent authenticating new EncryptionKey records, when a user already
        has an EncryptionKey. They should sign Paseto tokens with it, and perform
        key rotation with _existing_ key.
        """
        user, access_token = super().lookup_user_record(token, request)

        key_qs = user.encryption_keys.valid()
        if name := request.data.get('name') or request.POST.get('name'):
            key_qs = key_qs.filter(name=name)

        if key_qs.exists():
            raise exceptions.AuthenticationFailed(
                _("Token must be signed with one of the user's existing Encryption Keys!")
            )

        return VerifiedAuth(user, access_token)

    def load_public_key(self, request: Request) -> bytes:
        """
        Load public key from request Header, URL params or Body.

        The public key must be base64-encoded if it's in the URL params or Header.
        """
        param_key = get_setting('AUTO_AUTH_METHOD_PARAM')

        try:
            verify_public_key_bytes = request.data[param_key].encode()
        except (AttributeError, KeyError):
            verify_public_key_bytes = request.POST.get(param_key, '').encode()

        if not verify_public_key_bytes:
            if not (verify_public_key := request.META.get(f'HTTP_{param_key}'.upper(), '')):
                verify_public_key = request.GET.get(param_key, '')
            try:
                # Add 2 equals signs as padding to avoid "Incorrect padding" error. Unnecessary
                # padding is stripped, and 2 is the most ever required to avoid this specific error.
                verify_public_key_bytes = _base64.urlsafe_b64decode(f'{verify_public_key}==')
            except (binascii.Error, UnicodeDecodeError) as e:
                raise exceptions.AuthenticationFailed(
                    _('Unable to parse verification public key: %s') % e
                ) from None

        if not verify_public_key_bytes:
            raise exceptions.AuthenticationFailed(_('No verification key was provided'))

        return verify_public_key_bytes
