"""
These functions must be importable from other scripts
without requiring the Django environment to be setup!
"""
from lcconfig import ConsoleSettingsConfig, load_private_key
from lcconfig.encryption import mask
from pathlib import Path
from typing import NamedTuple, TYPE_CHECKING

if TYPE_CHECKING:
    from uuid import UUID

from .keys import PrivateKey, FacadePrivateKey

ENCRYPTION_SECTION = 'Encryption'
LOCAL_UMBRELLA_KEY_NAME = 'local_umbrella_key_cid'


class KeyDetails(NamedTuple):
    """
    Canonical ID and Key of Private Key to sign Paseto Tokens.
    """
    cid: 'UUID'
    user_cid: 'UUID'
    key: FacadePrivateKey


def get_key_password(cid, password_prefix='', ssh=False) -> bytes:
    """
    Return Private Key password with support for OpenSSH-formatted
    private keys which don't support passwords over 72 bytes.
    """
    try:
        cid = cid.hex
    except AttributeError:
        cid = str(cid).replace('-', '')

    # When bootstrapping consoles, lookup value in config file, in case
    # it's been generated just now as part of the initial setup process
    if not password_prefix:
        password_prefix = load_private_key('chacha')
        if not password_prefix:
            return b''

    password = f'{password_prefix}|{cid[::-1]}'.encode()

    if ssh and len(password) > 72:
        return password[-72:]

    return password


def load_pinned_verification_key(key_cid, user_cid, location) -> KeyDetails:
    """
    Portals have "pinned" key baked into the installation configuration file,
    used for registration purposes.

    Since it's included in the installation config, it cannot be secured via
    PRIVATE_KEY_PASSWORD_PREFIX machinery

    This key should _not_ be used after registration has been completed.
    """
    password = mask(key_cid).encode()

    try:
        pinned_key = Path(location)
        pk = PrivateKey.load_pem_from_file(pinned_key, password=password)
    except (FileNotFoundError, ValueError, TypeError):
        raise ValueError('Unable to load pinned verification key') from None

    return KeyDetails(key_cid, user_cid, pk)


def load_local_umbrella_api_key_from_file(user_cid, location) -> KeyDetails:
    """
    The local Encryption Key should be loadable from a file,
    so that scripts can use it without initializing the whole
    Portal to make a database call.
    """
    cfg = ConsoleSettingsConfig().as_typed_dict(ENCRYPTION_SECTION)
    if not (cid := cfg.get(LOCAL_UMBRELLA_KEY_NAME)):
        raise ValueError('Local verification key does not exist')

    password = get_key_password(cid, ssh=True)

    try:
        pinned_key = Path(location)
        pk = PrivateKey.load_pem_from_file(pinned_key, password=password)
    except (ValueError, ValueError, TypeError):
        raise ValueError('Unable to load local verification key') from None

    return KeyDetails(cid, user_cid, pk)
