from django.conf import settings
from django.db import models
from django.db.models import Q
from django.utils.translation import gettext_lazy as _
from django.utils.encoding import force_str, force_bytes

from typing import Optional

from console_base.theme import ICONS
from console_base.models import BaseUUIDPKModel, LCDateTimeField, LCTextField, NameCITextField
from console_base.models.fernet import EncryptedBinaryField

from .choices import STATUS
from .fields import KeyField
from .key_loaders import get_key_password
from .managers import EncryptionQuerySet
from .validators import validate_public_key
from . import keys


class EncryptionKey(BaseUUIDPKModel):
    """
    Store a public key or public/private key pair and associate it to a particular user.

    Implements the same concept as the OpenSSH ``~/.ssh/authorized_keys`` file on a Unix system.
    """
    css_icon = ICONS.TLS
    STATUS = STATUS

    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        verbose_name=_("User"),
        related_name="encryption_keys",
        on_delete=models.CASCADE,
    )

    # Key text in either PEM or OpenSSH format.
    key = KeyField(
        _("Encryption Key"),
        help_text=_("The user's RSA / Ed25519 / ECDSA public key"),
        validators=[validate_public_key],
    )
    kid = LCTextField(
        help_text=_('Paserk Key ID derived from the Public Key'),
    )

    # Private Key bytes in either PEM or OpenSSH format.
    # Just because this field is on the model doesn't mean we should use it.
    # Included here for fast backup / restores of cloud systems, which to
    # accomplish means we need to store the private key on disk anyway.
    private_key = EncryptedBinaryField(
        _("Private Key"),
        null=True,
        blank=True,
        help_text=_("The user's RSA / Ed25519 / ECDSA private key"),
    )

    # Info describing the key. What system is authenticating with the key etc.
    name = NameCITextField(_("Name"))

    # Date and time that key was last used for authenticating a request.
    last_used_on = LCDateTimeField(_("Last Used On"), null=True, blank=True)
    revocation_date = LCDateTimeField(_("Revocation Date"), null=True, blank=True)
    status = LCTextField(_("Status"), choices=STATUS.choices, default=STATUS.active)

    objects = EncryptionQuerySet.as_manager()

    class Meta:
        verbose_name = _("Encryption Key")
        verbose_name_plural = _("Encryption Keys")
        constraints = [
            models.UniqueConstraint(
                fields=('user', 'name'),
                name='unique_key_name_per_user',
                condition=Q(revocation_date__isnull=True),
            ),
            models.UniqueConstraint(
                fields=('key',),
                name='encryption_keys_must_be_unique',
            ),
            models.UniqueConstraint(
                fields=('kid',),
                name='public_key_ids_must_be_unique',
            ),
        ]

    def __str__(self):
        return f'{self.__class__.__name__}({self.name!r})'

    def get_key(self) -> keys.FacadePublicKey:
        key_bytes = force_bytes(self.key)
        exc, key = keys.PublicKey.load_serialized_public_key(key_bytes)
        if key is None:
            if exc is None:  # pragma: no cover
                raise ValueError("Failed to load key")
            raise exc
        return key

    def get_private_key(self) -> Optional[keys.FacadePrivateKey]:
        openssh_banner = b'-----BEGIN OPENSSH PRIVATE KEY-----'
        password = get_key_password(
            self.cid,
            settings.PRIVATE_KEY_PASSWORD_PREFIX,
            self.private_key.startswith(openssh_banner),
        )
        if not password:
            return None
        return keys.PrivateKey.load_pem(self.private_key, password=password)

    def save(self, *args, **kwargs) -> None:
        if not self.name:
            key_parts = force_str(self.key).split(" ")
            if len(key_parts) == 3:
                self.name = key_parts.pop()
        super().save(*args, **kwargs)
