import base64

from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings
from rest_framework import status
from rest_framework.test import APIClient

from ..models import EncryptionKey
from ..tokens import RegistrationToken, Token
from ..keys import Ed25519PrivateKey
from .. import get_setting

User = get_user_model()


class RestFrameworkPasetoTokenAuthenticationTest(TestCase):
    def setUp(self):
        self.api_client = APIClient(enforce_csrf_checks=True)
        self.user = User.objects.create_user(username="foo", is_superuser=True)

        self.key_ed25519 = Ed25519PrivateKey.generate()
        self.key_ed25519_orphan = Ed25519PrivateKey.generate()

        self.user_key_ed25519 = EncryptionKey.objects.create(
            name='ed255919-drf',
            user=self.user,
            key=self.key_ed25519.public_key.as_pem.decode(),
        )
        self.user_key_ed25519.refresh_from_db()

    def test_no_auth_unauthenticated_user(self):
        response = self.api_client.get(
            f'/api/encipher/encryptionkey/{self.user_key_ed25519.cid}/',
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_invalid_auth_unauthenticated_user(self):
        auth = 'Paseto what-a-crazy-token'
        response = self.api_client.get(
            f'/api/encipher/encryptionkey/{self.user_key_ed25519.cid}/',
            HTTP_AUTHORIZATION=auth,
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_unverifiable_key_unauthenticated_user(self):
        auth = 'Paseto eyJ0eXAiOiJKV1QiLCJhbGciOiJFZERTQSIsImtpZCI6IjJiNDIyZWNkZjM3YTYxNjVkOGRiMTY0MzE2N2UxNjEyYmZlNjE0ZmJmYWQwYmRiY2E2NjJkMGIyMDVlN2UxNGIifQ.eyJ1c2VybmFtZSI6ImZvbyIsInRpbWUiOjE2NDczNjg2MzksIm5vbmNlIjoibmNYTlpTaHlrWUkifQ.hNUay432cUdfUDYAfcRAvIpnWHPpmJK2saLdLvQAyllTDnzL9dpzehjv6uULOabTnzL_9YzulOnS4jXPh-tpCg'  # noqa
        response = self.api_client.get(
            f'/api/encipher/encryptionkey/{self.user_key_ed25519.cid}/',
            HTTP_AUTHORIZATION=auth,
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

    def test_signed_token_authenticates_user(self):
        auth = Token(user_cid=self.user.cid, private_key=self.key_ed25519).auth_header()
        self.assertTrue(self.user.has_perm('encipher.view_encryptionkey'))

        response = self.api_client.get(
            f'/api/encipher/encryptionkey/{self.user_key_ed25519.cid}/',
            HTTP_AUTHORIZATION=auth,
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    def test_auth_requesting_correct_public_key_succeeds(self):
        auth = Token(user_cid=self.user.cid, private_key=self.key_ed25519).auth_header()

        response = self.api_client.get(
            f'/api/encipher/encryptionkey/{self.user_key_ed25519.cid}/',
            HTTP_AUTHORIZATION=auth,
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.json())

    def test_auth_requesting_nonexistent_public_key_fails(self):
        auth = Token(
            user_cid=self.user.cid,
            private_key=self.key_ed25519_orphan,
        ).auth_header()

        response = self.api_client.get(
            f'/api/encipher/encryptionkey/{self.user_key_ed25519.cid}/',
            HTTP_AUTHORIZATION=auth,
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED, response.json())


class SignedTokenByProvidedKeyTest(TestCase):
    def setUp(self):
        self.api_client = APIClient(enforce_csrf_checks=True)
        self.user = User.objects.create_user(username="needs-encryption-key", is_superuser=True)
        self.key_ed25519 = Ed25519PrivateKey.generate()

    @override_settings(AUTO_REGISTER_ENCRYPTION_KEY_URL_NAMES={'api-encryptionkey-list'})
    def test_attached_public_key_authenticates_user(self):
        auth = RegistrationToken(user_cid=self.user.cid, private_key=self.key_ed25519).auth_header()
        self.assertTrue(self.user.has_perm('encipher.view_encryptionkey'))
        public_key = base64.urlsafe_b64encode(self.key_ed25519.public_key.as_pem).decode()

        response = self.api_client.head(
            f'/api/encipher/encryptionkey/?verify_public_key={public_key}',
            HTTP_AUTHORIZATION=auth,
        )
        self.assertEqual(response.status_code, status.HTTP_200_OK)

    @override_settings(AUTO_REGISTER_ENCRYPTION_KEY_URL_NAMES={'api-encryptionkey-list'})
    def test_posting_to_endpoint_with_valid_verification_key_succeeds(self):
        """
        Posting to API with token signed by value included in `verify_public_key`
        creates the EncryptionKey record.
        """
        user = User.objects.create_user(username="register-encryption-key", is_superuser=True)
        self.assertFalse(user.encryption_keys.exists(), msg='User should have NO encryption keys!')
        user.refresh_from_db()

        auth = RegistrationToken(user_cid=user.cid, private_key=self.key_ed25519).auth_header()
        param_key = get_setting('AUTO_AUTH_METHOD_PARAM')

        data = {
            'name': 'auto-register-key',
            'user': user.cid,
            'key': self.key_ed25519.public_key.as_pem.decode(),
            'kid': self.key_ed25519.public_key.as_paseto(version=4).to_paserk_id(),
            'revocation_date': '',
            param_key: self.key_ed25519.public_key.as_pem.decode(),
        }
        response = self.api_client.post(
            '/api/encipher/encryptionkey/',
            data=data,
            HTTP_AUTHORIZATION=auth,
        )
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
        self.assertTrue(user.encryption_keys.exists())

    @override_settings(AUTO_REGISTER_ENCRYPTION_KEY_URL_NAMES={'api-encryptionkey-list'})
    def test_posting_to_endpoint_with_preexisting_endpoint_fails(self):
        """
        Posting to API with token signed by value included in `verify_public_key`
        doesn't create EncryptionKey if user already has a EncryptionKey.
        """
        user = User.objects.create_user(username="register-encryption-key", is_superuser=True)
        existing = Ed25519PrivateKey.generate()
        EncryptionKey.objects.create(
            name='the-com-key',
            user=user,
            key=existing.public_key.as_pem.decode(),
        )

        auth = RegistrationToken(user_cid=user.cid, private_key=self.key_ed25519).auth_header()
        param_key = get_setting('AUTO_AUTH_METHOD_PARAM')

        data = {
            'name': 'the-com-key',
            'user': user.cid,
            'key': self.key_ed25519.public_key.as_pem.decode(),
            'kid': self.key_ed25519.public_key.as_paseto(version=4).to_paserk_id(),
            param_key: self.key_ed25519.public_key.as_pem.decode(),
        }
        response = self.api_client.post(
            '/api/encipher/encryptionkey/',
            data=data,
            HTTP_AUTHORIZATION=auth,
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED, response.json())

    @override_settings(AUTO_REGISTER_ENCRYPTION_KEY_URL_NAMES={'api-encryptionkey-list'})
    def test_posting_to_endpoint_with_invalid_verification_key_fails(self):
        """
        Posting to API with token that is NOT signed by value included in `verify_public_key`
        must be unauthorized and refuse to create EncryptionKey record.
        """
        user = User.objects.create_user(username="register-encryption-key", is_superuser=True)
        self.assertFalse(user.encryption_keys.exists(), msg='User should have NO public keys!')
        user.refresh_from_db()

        new_key = Ed25519PrivateKey.generate()
        auth = RegistrationToken(user_cid=self.user.cid, private_key=new_key).auth_header()
        param_key = get_setting('AUTO_AUTH_METHOD_PARAM')

        data = {
            'name': 'auto-register-key',
            'user': user.cid,
            'key': new_key.as_pem.decode(),
            param_key: self.key_ed25519.public_key.as_pem.decode(),
        }
        response = self.api_client.post(
            '/api/encipher/encryptionkey/',
            data=data,
            HTTP_AUTHORIZATION=auth,
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED, response.json())
        self.assertFalse(user.encryption_keys.exists())

    def test_attached_public_key_on_specified_endpoints(self):
        """
        Permit auto-registration of Encryption Keys only on specified endpoints.
        """
        user = User.objects.create_user(username="has-pk-otp-pk", is_superuser=True)
        existing = Ed25519PrivateKey.generate()
        EncryptionKey.objects.create(
            name='existing-key',
            user=user,
            key=existing.public_key.as_pem.decode(),
        )

        incorrectly_rotated = Ed25519PrivateKey.generate()
        auth = RegistrationToken(user_cid=user.cid, private_key=incorrectly_rotated).auth_header()
        self.assertTrue(user.has_perm('encipher.view_encryptionkey'))
        public_key = base64.urlsafe_b64encode(incorrectly_rotated.public_key.as_pem).decode()

        response = self.api_client.get(
            f'/api/encipher/encryptionkey/?verify_public_key={public_key}',
            HTTP_AUTHORIZATION=auth,
        )
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED, response.json())

        with override_settings(AUTO_REGISTER_ENCRYPTION_KEY_URL_NAMES={'api-encryptionkey-list'}):
            response = self.api_client.get(
                f'/api/encipher/encryptionkey/?verify_public_key={public_key}',
                HTTP_AUTHORIZATION=auth,
            )
            self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED, response.json())
