from datetime import datetime, timedelta, timezone
from freezegun import freeze_time
from lcconfig.otp import generate_totp
import json
from uuid import uuid1
import pyseto
from unittest.mock import patch
from django.contrib.auth import get_user_model
from django.test import TestCase, override_settings

from ..tokens import AccessToken, UnverifiedToken, RegistrationToken, Token
from .. import tokens, keys
from . import data

User = get_user_model()


class TokenTest(TestCase):
    def setUp(self):
        self.username = "rusty"
        self.user_cid = uuid1()
        self.private_key = keys.Ed25519PrivateKey.generate()

    @patch("secrets.token_urlsafe")
    def test_create_auth_header(self, mock_get_nonce):
        mock_get_nonce.return_value = "yVJ0MVWhqPQ"
        token = Token(private_key=self.private_key, user_cid=self.user_cid)
        header = token.auth_header()
        self.assertTrue(header.startswith("Paseto "))

        token_data = pyseto.decode(
            keys=self.private_key.as_paseto(version=4),
            token=f'{header.split(" ")[1]}==',
            deserializer=json,
        ).payload
        for k in ('exp', 'iat', 'nbf'):
            token_data.pop(k, None)

        self.assertDictEqual(
            token_data,
            {
                "nonce": "yVJ0MVWhqPQ",
                "user_cid": str(self.user_cid),
            },
        )

    @patch("secrets.token_urlsafe")
    def test_registration_auth_header(self, mock_get_nonce):
        mock_get_nonce.return_value = "yVJ0MVWhqPQ"
        token = RegistrationToken(private_key=self.private_key, user_cid=self.user_cid)
        header = token.auth_header()
        self.assertTrue(header.startswith("Auto_Paseto "))

        token_data = pyseto.decode(
            keys=self.private_key.as_paseto(version=4),
            token=f'{header.split(" ")[1]}==',
            deserializer=json,
        ).payload
        for k in ('exp', 'iat', 'nbf'):
            token_data.pop(k, None)

        self.assertEqual(
            token_data,
            {
                "nonce": "yVJ0MVWhqPQ",
                "user_cid": str(self.user_cid),
                "aud": "register",
            },
        )


class UnverifiedTokenTest(TestCase):

    def setUp(self):
        self.username = "rusty"
        self.user = User.objects.create(username=self.username, cid=uuid1())
        self.user_cid = self.user.cid
        self.user.refresh_from_db()
        self.private_key = keys.Ed25519PrivateKey.generate()
        self.pub_key = self.private_key.public_key
        self.kid = self.pub_key.as_paseto(version=4).to_paserk_id()
        self.token = Token(private_key=self.private_key, user_cid=self.user_cid)
        self.token_string = self.token.sign()

    def test_kid_footer(self):
        unverified_token = UnverifiedToken(self.token_string, self.pub_key.as_pem)
        footer = unverified_token.footer_payload()
        self.assertEqual(footer['kid'], self.kid)


class AccessTokenTest(TestCase):

    def setUp(self):
        self.username = "rusty"
        self.user = User.objects.create(username=self.username, cid=uuid1())
        self.user_cid = self.user.cid
        self.user.refresh_from_db()
        self.private_key = keys.Ed25519PrivateKey.generate()
        self.pub_key = self.private_key.public_key
        self.token = Token(private_key=self.private_key, user_cid=self.user_cid)
        self.token_string = self.token.sign()
        self.unverified_token = UnverifiedToken(self.token_string, self.pub_key.as_pem)
        self.access_token = self.unverified_token.verify()

    def test_get_claimed_user_cid(self):
        creds = self.access_token.get_claimed_creds()
        self.assertEqual(creds.get('user_cid'), str(self.user_cid))

    def test_verify_valid(self):
        token = self.access_token.verify()
        self.assertIsInstance(token, tokens.AccessToken)
        creds = self.access_token.get_claimed_creds()
        self.assertEqual(creds.get('user_cid'), str(self.user_cid))

    def test_verify_key_mismatch(self):
        pubkey = keys.PublicKey.load_pem(data.PEM_PUBLIC_RSA)
        with self.assertRaises(ValueError):
            UnverifiedToken(self.token_string, pubkey.as_pem).verify()

        private_key = keys.Ed25519PrivateKey.generate()
        pubkey = private_key.public_key
        with self.assertRaises(pyseto.VerifyError):
            UnverifiedToken(self.token_string, pubkey.as_pem).verify()

    def test_time_out_of_allowed_range_before(self):
        token = Token(private_key=self.private_key, user_cid=self.user_cid)
        token_string = token.sign()

        dt = datetime.now(tz=timezone.utc) - timedelta(seconds=300)
        with freeze_time(dt), self.assertRaises(pyseto.VerifyError):
            UnverifiedToken(token_string, self.pub_key.as_pem).verify()

    def test_time_out_of_allowed_range_after(self):
        token = Token(private_key=self.private_key, user_cid=self.user_cid)
        token_string = token.sign()

        dt = datetime.now(tz=timezone.utc) + timedelta(seconds=900)
        with freeze_time(dt), self.assertRaises(pyseto.VerifyError):
            UnverifiedToken(token_string, self.pub_key.as_pem).verify()

    def test_nonce_already_used(self):
        token1 = self.access_token.verify()
        self.assertIsInstance(token1, AccessToken)
        # Second attempt fails because nonce was already used
        with self.assertRaises(pyseto.VerifyError):
            self.access_token.verify()

    @override_settings(
        ENCIPHER=dict(
            NONCE_BACKEND="encipher.nonce.null.NullNonceBackend"
        )
    )
    def test_nonce_already_used_null_backend(self):
        token1 = self.access_token.verify()
        self.assertIsInstance(token1, AccessToken)
        # Second attempt works because null nonce backend doesn't do anything
        token2 = self.access_token.verify()
        self.assertIsInstance(token2, AccessToken)


class RegistrationTokenTest(TestCase):

    def setUp(self):
        self.username = "rusty"
        self.user = User.objects.create(username=self.username, cid=uuid1())
        self.user_cid = self.user.cid
        self.user.refresh_from_db()
        self.private_key = keys.Ed25519PrivateKey.generate()
        self.pub_key = self.private_key.public_key

    def test_otp_valid(self):
        public_key = self.private_key.public_key
        totp = generate_totp(public_key.as_pem.decode(), digits=8)

        token = RegistrationToken(private_key=self.private_key, user_cid=self.user_cid)
        token_string = token.sign()
        untrusted_token = UnverifiedToken(token_string, public_key.as_pem)
        access_token = untrusted_token.verify()
        self.assertIn('otp', access_token.footer)

        self.assertEqual(access_token.footer['otp'], totp.now())

        # Alter time so totp is invalid
        totp = generate_totp(public_key.as_pem.decode(), digits=8).now()
        dt = datetime.now(tz=timezone.utc) + timedelta(seconds=75)
        with freeze_time(dt):
            token = RegistrationToken(private_key=self.private_key, user_cid=self.user_cid)
            token_string = token.sign()
            token = UnverifiedToken(token_string, self.pub_key.as_pem).verify()

            self.assertIn('otp', token.footer)
            self.assertNotEqual(token.footer['otp'], totp)
