import base64

from django.contrib.auth import get_user_model
from django.core.cache import cache
from django.test import TestCase
from django.urls import reverse
from rest_framework import status
from rest_framework.test import APIClient

from .. import models
from ..keys import Ed25519PrivateKey
from ..tokens import Token
from encipher import get_setting
User = get_user_model()


class EncryptionKeyViewSetTest(TestCase):

    def setUp(self):
        cache.clear()
        self.user = User.objects.create_user(username="foo", is_superuser=True)
        self.private_key = Ed25519PrivateKey.generate()
        self.key_bytes = self.private_key.public_key.as_pem.strip()
        self.key_text = self.key_bytes.decode()
        self.public_key_obj = models.EncryptionKey.objects.create(
            name='ed255919-drf',
            user=self.user,
            key=self.key_text,
        )

    def test_encryption_key_list(self):
        api_client = APIClient()
        response = api_client.get(reverse("api-encryptionkey-list"))
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED)

        token = Token(user_cid=self.user.cid, private_key=self.private_key).auth_header()
        api_client.credentials(HTTP_AUTHORIZATION=token)
        response = api_client.get(reverse("api-encryptionkey-list"))
        self.assertEqual(response.status_code, status.HTTP_200_OK, response.data)

        key_data = response.data[0]
        self.assertEqual(key_data['key'], self.key_text)

    def test_create_public_key(self):
        api_client = APIClient()
        token = Token(user_cid=self.user.cid, private_key=self.private_key).auth_header()
        api_client.credentials(HTTP_AUTHORIZATION=token)

        new_private_key = Ed25519PrivateKey.generate()
        new_key_text = new_private_key.public_key.as_pem.strip().decode()

        key_data = {
            'name': f'new key for {self.user}',
            'user': str(self.user.cid),
            'key': new_key_text,
            'kid': new_private_key.public_key.as_paseto(version=4).to_paserk_id(),
            'revocation_date': '',
        }

        response = api_client.post(reverse("api-encryptionkey-list"), data=key_data)
        self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.data)
        self.assertEqual(response.data['key'], key_data['key'])

    def test_request_with_key_but_no_token(self):
        # Ensure that these invalid flows don't raise traceback
        api_client = APIClient()
        url = reverse("api-encryptionkey-list")
        param = get_setting('AUTO_AUTH_METHOD_PARAM')
        response = api_client.get(f'{url}?{param}={base64.urlsafe_b64encode(self.key_bytes)}')
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED, response.data)

        url = reverse("api-encryptionkey-detail", kwargs={'cid': self.public_key_obj.cid})
        param = get_setting('AUTO_AUTH_METHOD_PARAM')

        key_data = {
            param: self.key_text,
        }
        response = api_client.post(url, key_data)
        self.assertEqual(response.status_code, status.HTTP_401_UNAUTHORIZED, response.data)
