from datetime import datetime, timedelta, timezone
from lcconfig.otp import generate_totp, verify_totp
from lchttp.json import json_dumps, json_loads
from lclazy import LazyLoader
import json
import logging
from typing import NamedTuple, Optional, Union, TYPE_CHECKING
from uuid import UUID
if TYPE_CHECKING:
    from lcrequests.tokens import PasetoFooter
    import pyseto
    from pyseto import KeyInterface
else:
    pyseto = LazyLoader('pyseto', globals(), 'pyseto')

from django.conf import settings
from .models import EncryptionKey
from . import keys, get_setting
from .nonce import get_nonce_backend
from .typehints import ClaimCreds

logger = logging.getLogger(__name__)
_retrieve_key_from_aspen = None


# ---------------------------------------------------------------------------
class KeyData(NamedTuple):
    pk: UUID
    key: str
    last_used_on: datetime


# ---------------------------------------------------------------------------
class Token:
    """
    Represents a Paseto Token that's either been constructed
    by our code or has been verified to be valid.
    """
    auth_header_setting = 'AUTH_METHOD'
    version: int = 4

    def __init__(
            self,
            private_key: keys.PrivateKey,
            user_cid: Union[UUID, str] = '',
            email: str = '',
            expires: int = 90,
    ) -> None:
        if not user_cid and not email:
            raise ValueError('User Canonical ID or Email address must be provided')

        self.private_key = private_key
        self.public_key = private_key.public_key
        self.user_cid: str = str(user_cid)
        self.email: str = email
        self.expires = expires

    def auth_header(self) -> str:
        """
        Create an HTTP Authorization header.
        """
        return f"{get_setting(self.auth_header_setting)} {self.sign().decode()}"

    def payload(self) -> dict:
        """
        Return data for Paseto token payload.
        """
        now = datetime.now(tz=timezone.utc)
        nbf = now - timedelta(seconds=get_setting('TIMESTAMP_TOLERANCE'))
        exp = now + timedelta(seconds=self.expires)
        payload = {
            'nonce': get_nonce_backend().generate_nonce(),
            'nbf': nbf.isoformat(timespec="seconds"),
            'exp': exp.isoformat(timespec="seconds"),
        }

        if self.user_cid:
            payload['user_cid'] = self.user_cid
        elif self.email:
            payload['email'] = self.email

        return payload

    def footer(self) -> dict:
        return {'kid': self.public_key.as_paseto(self.version).to_paserk_id()}

    def sign(self) -> bytes:
        """
        Create and return signed authentication Paseto token
        """
        return pyseto.Paseto(
            exp=self.expires,
            include_iat=True,
            leeway=get_setting('TIMESTAMP_TOLERANCE'),
        ).encode(
            key=self.private_key.as_paseto(version=self.version),
            payload=json_dumps(self.payload()),
            footer=json_dumps(self.footer()),
            serializer=json,
        )


# ---------------------------------------------------------------------------
class V1Token(Token):
    version = 1

class V2Token(Token):
    version = 2

class V3Token(Token):
    version = 3

V4Token = Token


# ---------------------------------------------------------------------------
class RegistrationToken(Token):
    """
    This token must be used when auto-registering new EncryptionKeys.
    """
    auth_header_setting = 'AUTO_AUTH_METHOD'

    def footer(self) -> dict:
        """
        Add a one-time password value to provide some extra
        security against malicious registration attempts.
        """
        data = super().footer()
        data['otp'] = generate_totp(self.private_key.public_key.as_pem.decode(), digits=8).now()
        return data

    def payload(self) -> dict:
        data = super().payload()
        data['aud'] = 'register'
        return data


# ---------------------------------------------------------------------------
class UnverifiedToken:
    """
    Wrapper class for convenience methods to
    perform builtin token verification.
    """
    def __init__(self, token: Union[bytes, str], pem: Union[bytes, str] = ''):
        self.token_class = pyseto.Token.new(token)
        self.token_string = token
        self.pem: bytes = pem if isinstance(pem, bytes) else pem.encode()
        self._footer_payload = None
        self.key_data: Optional[KeyData] = None
        self.paseto_version = int((token if isinstance(token, str) else token.decode())[1])
        self._public_key: Optional['KeyInterface'] = None

    def verify(self) -> 'AccessToken':
        """
        Perform builtin Paseto verification and return the verified token
        """
        token = pyseto.Paseto(
            leeway=get_setting('TIMESTAMP_TOLERANCE'),
        ).decode(
            keys=self.paseto_public_key(),
            token=self.token_string,
            deserializer=json,
        )

        return AccessToken(token, self.pem, self.key_data)  # type: ignore

    def footer_payload(self) -> 'PasetoFooter':
        if self._footer_payload is None:
            try:
                self._footer_payload = json_loads(self.token_class.footer)
            except json.JSONDecodeError:
                logger.info(f'Unable to decode token footer: {self.token_class.footer}')
                self._footer_payload = {}
        return self._footer_payload

    def paseto_public_key(self) -> 'KeyInterface':
        if not self._public_key:
            if self.pem:
                key = self.pem
            else:
                self.key_data = EncryptionKey.objects.filter(
                    **self.verification_key_params(),
                ).values_list('pk', 'key', 'last_used_on', named=True).first()
                if not self.key_data and settings.CHECK_ASPEN_FOR_UNKNOWN_KEYS:
                    global _retrieve_key_from_aspen
                    if _retrieve_key_from_aspen is None:
                        from .aspen import retrieve_key_from_aspen as _retrieve_key_from_aspen
                    self.key_data = _retrieve_key_from_aspen(footer=self.footer_payload())
                if not self.key_data:
                    msg = 'Verification key not found in local database'
                    if settings.CHECK_ASPEN_FOR_UNKNOWN_KEYS:
                        msg = f'{msg} and cannot be retrieved from Aspen.'
                    raise pyseto.VerifyError(msg)
                key = self.key_data.key.encode()

            self._public_key = pyseto.Key.new(
                version=self.paseto_version,
                purpose='public',
                key=key,
            )

        return self._public_key

    def verification_key_params(self) -> dict:
        """
        Get the Key ID or Canonical ID of the key that signed this token.
        This token cannot be valid if it's not signed by the key with this ID.
        """
        footer = self.footer_payload()

        if cid := footer.get('cid'):
            return {'cid': cid}
        elif kid := footer.get('kid'):
            return {'kid': kid}

        raise pyseto.VerifyError('No Key ID or Key Canonical ID specified')


# ---------------------------------------------------------------------------
class UnverifiedRegistrationToken(UnverifiedToken):
    """
    Perform extra verification on tokens when doing registration.
    """

    def verify(self) -> 'AccessToken':
        access_token = super().verify()
        self.valid_otp()
        return access_token

    def valid_otp(self) -> None:
        """
        If the token contains a TOTP, it must be verified first.
        This step is primarily used for registering new EncryptionKey
        records automatically.
        """
        if not (otp_code := self.footer_payload().get('otp', '')):
            raise pyseto.VerifyError('Registration token must contain verification code')

        key_text = self.pem.decode() if isinstance(self.pem, bytes) else self.pem
        if not verify_totp(key_text, otp=otp_code):
            logger.error('TOTP: %s is invalid: %s', (otp_code, key_text))
            raise pyseto.VerifyError('Invalid TOTP value. Check your device clock.')


# ---------------------------------------------------------------------------
class AccessToken:
    """
    Represents a Paseto token received from user input (and not yet trusted)
    """

    def __init__(
            self,
            token: pyseto.Token,
            pem: bytes,
            key_data: KeyData,
    ) -> None:
        """
        :param token: Paseto claim
        """
        if isinstance(token.payload, dict):
            self.payload = token.payload
        else:
            raise pyseto.VerifyError('Token payload must be data object.')

        self.token = token
        self.footer = token.footer
        self.pem = pem
        self.key_data = key_data

    def verify(self) -> 'AccessToken':
        """
        Verify the validity of the given Paseto Token using the given public key.
        """
        expiration_ts = self.validate_exp()
        claimed_nonce = self.validate_nonce()
        self.validate_unique_nonce(expiration_ts, claimed_nonce)
        self.update_last_used_datetime()

        return self

    def update_last_used_datetime(self) -> None:
        """
        Update last_used_on date if it hasn't been updated recently.
        Not updating every single time, since a token might be
        used many times per minute.
        """
        try:
            last_used_on = self.key_data.last_used_on
            now = datetime.now(tz=timezone.utc)
        except AttributeError:
            return

        if (
                not last_used_on
                or now - last_used_on > timedelta(minutes=get_setting('LAST_USED_INTERVAL'))
        ):
            EncryptionKey.objects.filter(pk=self.key_data.pk).update(last_used_on=now)

    def get_claimed_creds(self) -> ClaimCreds:
        """
        After verifying that the token is valid, get the User record
        from the token payload.
        """
        return ClaimCreds(
            user_cid=self.payload.get('user_cid', ''),
            email=self.payload.get('email', ''),
        )

    def validate_exp(self) -> int:
        """
        By the time this code has been reached, we know that the 'exp' value
        has passed builtin validation, but we _require_ all tokens to have
        the `exp` key present in the payload.
        """
        try:
            return int(datetime.fromisoformat(self.payload['exp']).timestamp())
        except (KeyError, ValueError):
            raise pyseto.VerifyError('Expiration value is required') from None

    def validate_nonce(self) -> str:
        """
        The payload must contain a nonce value!
        """
        try:
            return self.payload['nonce']
        except KeyError:
            raise pyseto.VerifyError('Nonce value is required.') from None

    def validate_unique_nonce(self, expiration_ts: int, nonce: str) -> None:
        """
        Ensure that nonce is unique to prevent replay attacks.
        """
        claimed_time = int(expiration_ts - datetime.utcnow().timestamp())

        nonce_backend = get_nonce_backend()
        if not nonce_backend.validate_nonce(claimed_time, nonce):
            logger.error("Claimed nonce has already been used - payload: %s", nonce)
            raise pyseto.VerifyError('Nonce has already been used.')

        # If we've gotten this far, the token is valid
        nonce_backend.log_used_nonce(claimed_time, nonce)
